"""
osv to NVD CVE converter

This module fetches the vulnerability data from osv.dev and stores them in NVD CVE 1.1 json format.
"""

from os import getenv
from zipfile import ZipFile

import httpx
import orjson
from semver import Version

from vdb.lib import CustomNamedTemporaryFile, config
from vdb.lib.nvd import NvdSource
from vdb.lib.utils import (
    compress_str,
    convert_score_severity,
    get_cvss3_from_vector,
    get_cvss4_from_vector,
    get_default_cve_data,
    parse_purl,
)
from vdb.lib.gha import GitHubSource

OSV_PER_VERSION_LIMIT = 10

# Size of the stream to read and write to the file
DOWNLOAD_CHUNK_SIZE = 4096

vendor_overrides = {
    "apk": "alpine",
    "deb": "debian",
    "go": "golang",
    "crates.io": "cargo",
    "swifturl": "swift",
    "github actions": "github",
}


def functions_to_modules(affected_functions):
    affected_modules = set()
    for af in affected_functions:
        if "::" in af:
            tmp_parts = af.split("::")
            affected_modules.add("::".join(tmp_parts[:-1]))
        elif "/" in af:
            tmp_parts = af.split("/")
            affected_modules.add("/".join(tmp_parts[:-1]))
    return affected_modules


class OSVSource(NvdSource):
    """OSV CVE source"""

    def download_all(self):
        """Download all cve data"""
        # For performance do not retain the whole data in-memory
        # See: https://github.com/AppThreat/vulnerability-db/issues/27
        data_list = []
        for _, url in config.OSV_URL_DICT.items():
            data = self.fetch(url)
            if data:
                self.store(data)
        return data_list

    def download_recent(self):
        pass

    def fetch(self, url):
        ret_data = []
        client = httpx.Client(http2=True, follow_redirects=True, timeout=180)
        with CustomNamedTemporaryFile() as tf:
            try:
                with client.stream("GET", url) as r:
                    for chunk in r.iter_bytes(chunk_size=DOWNLOAD_CHUNK_SIZE):
                        tf.write(chunk)
                    tf.flush()
            except Exception:
                return ret_data
            with ZipFile(tf.name, "r") as zipfp:
                for zf in zipfp.namelist():
                    if zf.endswith(".json"):
                        with zipfp.open(zf) as jsonf:
                            cve_data = jsonf.read()
                            try:
                                json_data = orjson.loads(cve_data)
                                ret_data += self.convert(json_data)
                            except Exception:
                                pass
        return ret_data

    def convert(self, cve_data):
        if cve_data.get("withdrawn"):
            return []
        return self.to_vuln(cve_data)

    @staticmethod
    def to_vuln(cve_data):
        ret_data = []
        if not cve_data.get("affected"):
            return ret_data
        cve_id = cve_data.get("id")
        # Remove unwanted DEBIAN- and ALPINE- prefixes
        if "CVE-" in cve_id and not cve_id.startswith("CVE-"):
            cve_id = cve_id[cve_id.find("CVE-") :]
        cwe_id = ""
        edition = "*"
        cve_references = cve_data.get("references", [])
        aliases = cve_data.get("aliases", [])
        # Related list in OSV is full of self-promoting ids from commercial companies
        related = [f for r in cve_data.get("related", []) if r.startswith("CVE-") or r.startswith("GHSA-")]
        aliases_block = ""
        if aliases and len(aliases) > 1:
            aliases_block = f"""
## Aliases
{", ".join(aliases)}
"""
        if related:
            aliases_block += f"""
## Related
{", ".join(related)}
"""
        description = f"""# {cve_data.get("summary", "Summary")}
{cve_data.get("details", "")}
{aliases_block}
"""
        if "** DISPUTED **" in description or "** REJECT **" in description:
            return ret_data
        references = []
        # Change the key from type to name in references
        for aref in cve_references:
            references.append({"name": aref.get("type", "url"), "url": aref.get("url")})
        # Red Hat can have a lot of references for some reason
        if len(references) > 512:
            references = references[:512]
        references = orjson.dumps(references, option=orjson.OPT_NAIVE_UTC)
        if isinstance(references, bytes):
            references = references.decode("utf-8", "ignore")
        # Offer an option to ignore malware data to keep the db size small
        if getenv("OSV_EXCLUDE_MALWARE") and cve_id.startswith("MAL"):
            return ret_data
        # Quality of PYSEC and GO data are quite low missing both severity and score
        # Where a PYSEC or GO feed also reference a github id, let's ignore it
        # Fixes #99
        if cve_id.startswith("PYSEC") or cve_id.startswith("GO-"):
            for i in aliases:
                if i.startswith("GHSA"):
                    return ret_data
        # Try to locate the CVE id from the aliases section
        if not cve_id.startswith("CVE") and not cve_id.startswith("RUSTSEC"):
            for i in aliases:
                if i.startswith("CVE"):
                    cve_id = i
                    break
        assigner = "google"
        root_vector_string = ""
        root_cvss4_vector_string = ""
        root_severity = "LOW"
        if cve_id.startswith("GHSA"):
            assigner = "github_m"
        elif cve_id.startswith("CVE"):
            assigner = "mitre"
        elif cve_id.startswith("NPM"):
            assigner = "github_m"
        # For malwares, default to critical
        if cve_id.startswith("MAL"):
            root_severity = "CRITICAL"

        if cve_data.get("severity"):
            severity_list = cve_data.get("severity")
            for sv in severity_list:
                if sv["type"] == "CVSS_V3":
                    root_vector_string = sv["score"]
                elif sv["type"] == "CVSS_V4":
                    root_cvss4_vector_string = sv["score"]
                elif sv["type"] == "CVSS_V2" and not root_vector_string:
                    root_vector_string = sv["score"]

        cve_database_specific = cve_data.get("database_specific")
        cve_ecosystem_specific = cve_data.get("ecosystem_specific")
        # Support for collecting affected functions
        affected_functions = set()
        affected_modules = set()
        if cve_database_specific:
            if cve_database_specific.get("severity"):
                root_severity = cve_database_specific.get("severity")
            if cve_database_specific.get("cwe_ids"):
                cwes = cve_database_specific.get("cwe_ids")
                if isinstance(cwes, list):
                    cwe_id = ",".join(cwes)
        if cve_ecosystem_specific and cve_ecosystem_specific.get("severity"):
            root_severity = cve_ecosystem_specific.get("severity")
        for pkg_data in cve_data.get("affected"):
            min_ver = None
            max_ver = None
            # Database specific values. Useful to override.
            ds_version_start_including = ""
            ds_version_end_including = ""
            ds_version_start_excluding = ""
            ds_version_end_excluding = ""
            severity = root_severity
            vector_string = root_vector_string
            cvss4_vector_string = root_cvss4_vector_string
            if pkg_data.get("severity"):
                for sv in pkg_data.get("severity"):
                    if sv["type"] == "CVSS_V4":
                        cvss4_vector_string = sv["score"]
                    elif sv["type"] == "CVSS_V3":
                        vector_string = sv["score"]
                    elif sv["type"] == "CVSS_V2" and not vector_string:
                        vector_string = sv["score"]
            if pkg_data.get("ecosystem_specific"):
                ecosystem_specific = pkg_data.get("ecosystem_specific")
                if ecosystem_specific.get("severity"):
                    severity = ecosystem_specific.get("severity")
                if ecosystem_specific.get("affected_functions"):
                    affected_functions = affected_functions.union(
                        ecosystem_specific.get("affected_functions")
                    )
                if ecosystem_specific.get("affects", {}).get("functions"):
                    affected_functions = affected_functions.union(
                        ecosystem_specific.get("affects").get("functions")
                    )
                for aimp in ecosystem_specific.get("imports", []):
                    if aimp.get("path"):
                        affected_modules.add(aimp.get("path"))
                    if aimp.get("symbols"):
                        affected_functions = affected_functions.union(
                            aimp.get("symbols")
                        )
            if pkg_data.get("database_specific"):
                database_specific = pkg_data.get("database_specific")
                if database_specific.get("cwes"):
                    cwes = database_specific.get("cwes")
                    if isinstance(cwes, list):
                        cwe_id = cwes[0].get("cweId")
                if database_specific.get("cvss"):
                    cvss = database_specific.get("cvss")
                    if isinstance(cvss, dict):
                        if cvss.get("severity"):
                            severity = cvss.get("severity", "").upper()
                        if not vector_string and cvss.get("vectorString"):
                            vector_string = cvss.get("vectorString")
                        if cvss.get("score"):
                            score = cvss.get("score")
                            severity = convert_score_severity(score)
                if database_specific.get("last_known_affected_version_range"):
                    (
                        ds_version_start_including,
                        ds_version_end_including,
                        ds_version_start_excluding,
                        ds_version_end_excluding,
                    ) = GitHubSource.get_version_range(
                        database_specific.get("last_known_affected_version_range")
                    )
                    min_ver = ds_version_start_including
                    max_ver = ds_version_end_including
            user_interaction = "REQUIRED"
            score = 0.0
            exploitability_score = 0.0
            attack_complexity = "LOW"
            # We give importance to cvss4 which always seems to be lower compared to cvss3
            if cvss4_vector_string:
                cvss4_obj = get_cvss4_from_vector(cvss4_vector_string)
                score = cvss4_obj.get("baseScore")
                severity = cvss4_obj.get("baseSeverity")
                exploitability_score = score
                attack_complexity = cvss4_obj.get("attackComplexity")
                # CVSS 3.1 only supports NONE or REQUIRED for userInteraction
                user_interaction = (
                    "REQUIRED"
                    if cvss4_obj.get("userInteraction", "") != "NONE"
                    else "NONE"
                )
            if vector_string:
                cvss3_obj = get_cvss3_from_vector(vector_string)
                # Fallback to CVSS 3.1 only if CVSS 4 is absent
                if not cvss4_vector_string:
                    score = cvss3_obj.get("baseScore")
                    severity = cvss3_obj.get("baseSeverity")
                    exploitability_score = cvss3_obj.get("temporalScore")
                    attack_complexity = cvss3_obj.get("attackComplexity")
                    user_interaction = cvss3_obj.get("userInteraction")
            if not vector_string and not cvss4_vector_string:
                (
                    score,
                    severity,
                    dvector_string,
                    attack_complexity,
                ) = get_default_cve_data(severity)
                # Override the score for malware
                if cve_id.startswith("MAL"):
                    score = 10.0
                    user_interaction = "NONE"
                # Set the default vector string only if unavailable
                if not vector_string and dvector_string:
                    vector_string = dvector_string
                exploitability_score = score
            ranges = pkg_data.get("ranges", [])
            versions_list = pkg_data.get("versions", [])
            vendor_ecosystem = pkg_data.get("package", {}).get("ecosystem", "").lower()
            vendor = vendor_ecosystem
            pkg_name = pkg_data.get("package", {}).get("name", "")
            pkg_name_list = []
            purl_prefix = pkg_data.get("package", {}).get("purl", "")
            purl = parse_purl(purl_prefix)
            if purl and vendor_ecosystem != "npm":
                if purl.get("type"):
                    vendor = purl["type"]
                if purl.get("namespace") and purl.get("name"):
                    pkg_name = f"{purl['namespace']}/{purl['name']}"
                elif purl.get("name"):
                    pkg_name = purl["name"]
            if ":" in pkg_name and vendor.lower() not in (
                "swift",
                "swifturl",
                "github",
                "github actions",
            ):
                # Example: commons-fileupload:commons-fileupload
                # org.apache.tomcat:tomcat
                tmp_a = pkg_name.split(":")
                if len(tmp_a) == 2:
                    pkg_name = f"{tmp_a[0]}/{tmp_a[-1]}"
            # In case of swift, we need to remove any https protocol from the name
            if pkg_name.startswith("https://"):
                pkg_name = pkg_name.removeprefix("https://")
            if pkg_name:
                pkg_name_list.append(pkg_name)
            # Substitute alpine for apk and debian for deb
            if vendor_overrides.get(vendor):
                vendor = vendor_overrides.get(vendor)
            # Since swift allows both url and local based lookups, we store both the variations
            if vendor in (
                "swift",
                "swifturl",
                "github",
                "github actions",
            ) and pkg_name.startswith("github.com"):
                pkg_name_list.append(pkg_name.removeprefix("github.com/"))
            # For OS packages, such as alpine OSV appends the os version to the vendor
            # Let's remove it and add it to package name
            if ":" in vendor_ecosystem and (
                "alpine" in vendor
                or "apk" in vendor
                or "deb" in vendor
                or "debian" in vendor
                or "almalinux" in vendor
                or "rocky" in vendor
            ):
                tmp_v = vendor_ecosystem.split(":")
                vendor = tmp_v[0].lower().replace(" ", "").replace("-", "")
                vdistro = tmp_v[1]
                if vendor in ("alpine", "apk"):
                    vdistro = vdistro.replace("v", "")
                # In os-release, ID for rockylinux is simply rocky
                if "rocky" in vendor:
                    vendor = vendor.replace("linux", "")
                edition = f"{vendor}-{vdistro}"
                # Only use the precise version for os packages
                if (
                    "debian" in vendor
                    or "deb" in vendor
                    or "alpine" in vendor
                    or "apk" in vendor
                    or "almalinux" in vendor
                    or "rocky" in vendor
                ):
                    pkg_name_list = [
                        f"{vendor}/{edition}/{pkg_name.removeprefix(vendor + '/')}"
                    ]
                else:
                    pkg_name_list.append(f"{edition}/{pkg_name}")
            # For some ecosystem, osv provides a full list of versions with partial events. See osv-pypi2.json for an example
            needs_version_backup = len(ranges) == 0
            possible_fix_version = ""
            for r in ranges:
                events = r.get("events")
                for ev in events:
                    if ev.get("introduced", "") in (0, "0", "0.0.0"):
                        needs_version_backup = True
                    if ev.get("fixed"):
                        possible_fix_version = ev.get("fixed")
                    if ev.get("limit") and not possible_fix_version:
                        possible_fix_version = ev.get("limit")

            if needs_version_backup and len(versions_list):
                # For smaller lists, we track them all
                if len(versions_list) < OSV_PER_VERSION_LIMIT:
                    for each_version in versions_list:
                        for full_pkg in pkg_name_list:
                            tdata = config.CVE_TPL % dict(
                                cve_id=cve_id,
                                cwe_id=cwe_id,
                                assigner=assigner,
                                references=references,
                                description="",
                                vectorString=vector_string,
                                vendor=vendor,
                                product=full_pkg,
                                version="*",
                                edition=edition,
                                version_start_including=each_version,
                                version_end_including=each_version,
                                version_start_excluding="",
                                version_end_excluding="",
                                fix_version_start_including=possible_fix_version,
                                fix_version_end_including=possible_fix_version,
                                fix_version_start_excluding="",
                                fix_version_end_excluding="",
                                severity=severity,
                                attackComplexity=attack_complexity,
                                score=score,
                                userInteraction=user_interaction,
                                exploitabilityScore=exploitability_score,
                                publishedDate=cve_data.get("published"),
                                lastModifiedDate=cve_data.get("modified"),
                            )
                            try:
                                vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                                if vuln is None:
                                    continue
                                vuln.description = compress_str(description)
                                if affected_functions:
                                    vuln.affects = {
                                        "affected_functions": sorted(list(affected_functions)),
                                        "affected_modules": sorted(
                                            list(
                                                affected_modules
                                                | functions_to_modules(affected_functions)
                                            )
                                        ),
                                    }
                                if cvss4_vector_string:
                                    vuln.cvss4_vector_string = cvss4_vector_string
                                if purl_prefix:
                                    vuln.purl_prefix = purl_prefix
                                ret_data.append(vuln)
                            except Exception:
                                pass
                else:
                    try:
                        min_ver = min(
                            versions_list,
                            key=lambda x: Version.parse(x, optional_minor_and_patch=True),
                        )
                        max_ver = max(
                            versions_list,
                            key=lambda x: Version.parse(x, optional_minor_and_patch=True),
                        )
                    except Exception:
                        min_ver = versions_list[0]
                        max_ver = versions_list[-1]
                    for full_pkg in pkg_name_list:
                        tdata = config.CVE_TPL % dict(
                            cve_id=cve_id,
                            cwe_id=cwe_id,
                            assigner=assigner,
                            references=references,
                            description="",
                            vectorString=vector_string,
                            vendor=vendor,
                            product=full_pkg,
                            version="*",
                            edition=edition,
                            version_start_including=min_ver,
                            version_end_including=max_ver,
                            version_start_excluding="",
                            version_end_excluding=possible_fix_version,
                            fix_version_start_including=possible_fix_version,
                            fix_version_end_including=possible_fix_version,
                            fix_version_start_excluding=max_ver,
                            fix_version_end_excluding="",
                            severity=severity,
                            attackComplexity=attack_complexity,
                            score=score,
                            userInteraction=user_interaction,
                            exploitabilityScore=exploitability_score,
                            publishedDate=cve_data.get("published"),
                            lastModifiedDate=cve_data.get("modified"),
                        )
                        try:
                            vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                            if vuln is None:
                                continue
                            vuln.description = compress_str(description)
                            if affected_functions:
                                vuln.affects = {
                                    "affected_functions": sorted(list(affected_functions)),
                                    "affected_modules": sorted(
                                        list(
                                            affected_modules
                                            | functions_to_modules(affected_functions)
                                        )
                                    ),
                                }
                            if cvss4_vector_string:
                                vuln.cvss4_vector_string = cvss4_vector_string
                            if purl_prefix:
                                vuln.purl_prefix = purl_prefix
                            ret_data.append(vuln)
                        except Exception:
                            pass
            # Normal flow with version ranges
            for r in ranges:
                if r.get("type") == "GIT" and r.get("repo"):
                    vendor = "generic"
                    repo_name = (
                        r.get("repo")
                        .removeprefix("http://")
                        .removeprefix("https://")
                        .removeprefix("git://")
                        .removesuffix("/")
                        .removesuffix(".git")
                        .lower()
                    )
                    # See #112
                    for special_type in ("github.com", "gitlab.com"):
                        if repo_name.startswith(special_type):
                            vendor = special_type.removesuffix(".com")
                            repo_name = repo_name.replace(f"{special_type}/", "")
                    pkg_name_list.append(repo_name)
                events = r.get("events")
                rversions_list = r.get("versions", [])
                version_start_including = ds_version_start_including or min_ver
                version_end_including = ds_version_end_including or max_ver
                version_start_excluding = ds_version_start_excluding
                version_end_excluding = ds_version_end_excluding
                fix_version_start_including = ""
                fix_version_end_including = ""
                fix_version_start_excluding = ""
                fix_version_end_excluding = ""

                if rversions_list:
                    version_start_including = rversions_list[0]
                    if (
                        len(rversions_list) > 1
                        and version_start_including != rversions_list[-1]
                    ):
                        version_end_including = rversions_list[-1]
                fixed_version_in_events = (
                    len(
                        [
                            e
                            for e in events
                            if e.get("fixed")
                            or e.get("last_affected")
                            or e.get("limit")
                        ]
                    )
                    > 0
                )
                for ev in events:
                    # Reset all versions for introduced event
                    if ev.get("introduced") is not None:
                        version_end_including = ds_version_end_including
                        version_start_excluding = ds_version_start_excluding
                        version_end_excluding = ds_version_end_excluding
                        fix_version_start_including = version_end_excluding
                        fix_version_end_including = ""
                        fix_version_start_excluding = ""
                        fix_version_end_excluding = ""
                        intr = ev.get("introduced")
                        if (
                            ":" in str(intr)
                            and "debian" not in vendor
                            and "alpine" not in vendor
                        ):
                            version_start_including = str(intr).split(":")[-1]
                        else:
                            version_start_including = str(intr)

                        if version_start_including in (0, "0", "0.0.0"):
                            if min_ver:
                                version_start_including = min_ver
                            else:
                                version_start_including = "0.0.0"
                    if ev.get("fixed"):
                        fixed_val = ev.get("fixed")
                        if (
                            ":" in str(fixed_val)
                            and "debian" not in vendor
                            and "alpine" not in vendor
                        ):
                            version_end_excluding = str(fixed_val).split(":")[-1]
                        else:
                            version_end_excluding = str(fixed_val)
                        fix_version_start_including = version_end_excluding
                        fix_version_end_including = version_end_excluding
                        if not version_end_including and max_ver:
                            version_end_including = max_ver
                    if ev.get("last_affected"):
                        la_val = ev.get("last_affected")
                        if (
                            ":" in str(la_val)
                            and "debian" not in vendor
                            and "alpine" not in vendor
                        ):
                            version_end_including = str(la_val).split(":")[-1]
                        else:
                            version_end_including = str(la_val)
                        fix_version_start_excluding = version_end_including
                    if ev.get("limit"):
                        limit_val = ev.get("limit")
                        if (
                            ":" in str(limit_val)
                            and "debian" not in vendor
                            and "alpine" not in vendor
                        ):
                            version_end_excluding = str(limit_val).split(":")[-1]
                        else:
                            version_end_excluding = str(limit_val)
                    if version_start_including and (
                        (fix_version_start_including or fix_version_start_excluding)
                        or (not fixed_version_in_events and version_end_including)
                        or version_end_excluding
                        or (not fixed_version_in_events and not versions_list)
                    ):
                        for full_pkg in pkg_name_list:
                            tdata = config.CVE_TPL % dict(
                                cve_id=cve_id,
                                cwe_id=cwe_id,
                                assigner=assigner,
                                references=references,
                                description="",
                                vectorString=vector_string,
                                vendor=vendor,
                                product=full_pkg,
                                version="*",
                                edition=edition,
                                version_start_including=version_start_including,
                                version_end_including=version_end_including,
                                version_start_excluding=version_start_excluding,
                                version_end_excluding=version_end_excluding,
                                fix_version_start_including=fix_version_start_including,
                                fix_version_end_including=fix_version_end_including,
                                fix_version_start_excluding=fix_version_start_excluding,
                                fix_version_end_excluding=fix_version_end_excluding,
                                severity=severity,
                                attackComplexity=attack_complexity,
                                score=score,
                                userInteraction=user_interaction,
                                exploitabilityScore=exploitability_score,
                                publishedDate=cve_data.get("published"),
                                lastModifiedDate=cve_data.get("modified"),
                            )
                            try:
                                vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                                if vuln is None:
                                    continue
                                vuln.description = compress_str(description)
                                if affected_functions:
                                    vuln.affects = {
                                        "affected_functions": sorted(
                                            affected_functions
                                        ),
                                        "affected_modules": sorted(
                                            list(
                                                affected_modules
                                                | functions_to_modules(
                                                    affected_functions
                                                )
                                            )
                                        ),
                                    }
                                if cvss4_vector_string:
                                    vuln.cvss4_vector_string = cvss4_vector_string
                                if purl_prefix:
                                    vuln.purl_prefix = purl_prefix
                                ret_data.append(vuln)
                            except Exception:
                                pass
        return ret_data



import datetime
import gzip
import logging
from collections import defaultdict
from urllib.parse import urlparse

import httpx
import orjson

from vdb.lib import (
    CPE_FULL_REGEX,
    CustomNamedTemporaryFile,
    CvssV3,
    Vulnerability,
    VulnerabilityDetail,
    config,
)
from vdb.lib.cve import CVESource
from vdb.lib.utils import clean_cpe_uri, extract_affected_symbols, url_to_purl


logging.basicConfig(
    level=logging.INFO, format="%(levelname)s [%(asctime)s] %(message)s"
)
LOG = logging.getLogger(__name__)

# Start year of historic cve data
start_year = config.NVD_START_YEAR
# Current time
now = datetime.datetime.now()
# Size of the stream to read and write to the file
DOWNLOAD_CHUNK_SIZE = 128

cpe_proposal_cache = defaultdict(list)


def filterable_cve(cve_id: str) -> bool:
    """Method to check if the CVE id is less than the start year"""
    if not cve_id.startswith("CVE"):
        return False
    for year in range(1999, start_year):
        if cve_id.startswith(f"CVE-{year}-"):
            return True
    return False


def get_version(inc_version: str, exc_version: str) -> str:
    """
    Method to determine whether to use including or excluding version
    :param inc_version: Including version
    :param exc_version: Excluding version
    :return: Including version if it is not empty or *. Excluding version otherwise
    """
    if inc_version and inc_version != "*":
        return inc_version
    return exc_version


def filterable_git_url(url: str, hostname: str) -> bool:
    if "git" not in hostname:
        return True
    for part in (
        "cve",
        "disclosure",
        "secdb",
        "research",
        "exploit",
        "security-advisory",
        "advisories",
        "bulletins",
        "pocs",
        "_poc",
        "/poc",
        "0day",
        "vulnerabilit",
        "xss",
        "cisagov",
        "-post/",
        "_posts",
        ".pdf",
        "covering360",
        "fuzz",
        "-csrf",
        "advisory-db",
        "defcon",
        "audit-",
        "announcements",
        "divide-by-zero",
        "security-research",
        "apidoc",
        "-query-help",
        "/blog",
        "/news",
        "/support/",
        "/bug_report",
        "nu11secur1ty",
    ):
        if part in url.lower():
            return True
    for part in (
        "github.io",
        "gist.github.com",
        "about.gitlab.com",
        "lists.apache.org",
        "gitbooks.io",
        "githubusercontent.com",
        "enterprise.github.com",
        "git-scm.com",
        "docs.",
        "github.blog",
        "github.dev",
    ):
        if part in hostname.lower():
            return True
    return False


def get_alt_cpes(cpe_uri, git_urls):
    alt_cpes = []
    parsed_git_repo_names: dict[str, bool] = {}
    # Try to extract any git references from related urls
    # See: https://github.com/AppThreat/vulnerability-db/issues/91
    for agit_url in git_urls:
        url_obj = urlparse(agit_url)
        # Ignore obvious filterable urls
        if filterable_git_url(agit_url, url_obj.hostname) or (
            not url_obj.path and not url_obj.query
        ):
            continue
        purl_obj = url_to_purl(agit_url)
        if not purl_obj:
            continue
        git_repo_product = f"{purl_obj.get('namespace')}/{purl_obj.get('name')}"
        if not parsed_git_repo_names.get(git_repo_product):
            parsed_git_repo_names[git_repo_product] = True
            purl_type = purl_obj["type"]
            # Tune down the CPE proposal to reduce false positives
            # See issue #224
            if cpe_uri.startswith(f"cpe:2.3:a:{purl_type}:") or (
                purl_type in cpe_uri
                and "git" not in purl_type
                and "git" not in git_repo_product
            ):
                # We only need 2 new aliases
                if len(cpe_proposal_cache.get(cpe_uri, [])) > 2:
                    cpe_proposal_cache[cpe_uri].pop(0)
                cpe_proposal_cache[cpe_uri].append(
                    f"cpe:2.3:a:{purl_type}:{git_repo_product}:*:*:*:*:*:*:*:*"
                )
    # See if there is something useful in the cache
    if not alt_cpes:
        alt_cpes = cpe_proposal_cache.get(cpe_uri, [])
    return alt_cpes


class NvdSource(CVESource):
    """NVD CVE source. This uses CVE json 1.1 format that are split based on the year"""

    def download_all(self):
        """Download all historic cve data"""
        super().download_all()
        for y in range(now.year, int(start_year) - 1, -1):
            data = self.fetch(y)
            if not data:
                continue
            self.store(data)

    def download_recent(self):
        """Method which downloads the recent CVE gzip from NVD"""
        # Create database
        data = self.fetch("recent")
        if data:
            self.store(data)
        return data

    def fetch(self, year: int | str) -> list[Vulnerability] | None:
        """Private Method which downloads the given CVE gzip from NVD"""
        url = config.NVD_URL % dict(year=year)
        LOG.debug("Download NVD CVE from %s", url)
        with CustomNamedTemporaryFile() as tf:
            try:
                with httpx.stream("GET", url, follow_redirects=True, timeout=180) as r:
                    for chunk in r.iter_bytes(chunk_size=DOWNLOAD_CHUNK_SIZE):
                        tf.write(chunk)
                    tf.flush()
            except Exception:
                logging.warning("Exception while downloading NVD feed from %s", url)
                return None
            with gzip.open(tf.name, "rb") as gzipjf:
                try:
                    cve_data = gzipjf.read()
                    json_data = orjson.loads(cve_data)
                    return self.convert(json_data)
                except Exception:
                    logging.warning(
                        "Exception while parsing NVD CVE feed for %s. Please try after some time",
                        year,
                    )
                    return None

    def convert(self, cve_data: dict) -> list[Vulnerability]:
        """Convert cve data to Vulnerability"""
        # If this the data from the api, use the new convert_api method
        if not cve_data.get("CVE_Items") and cve_data.get("vulnStatus"):
            return self._convert_api(cve_data)
        items = cve_data.get("CVE_Items")
        data = []
        for cve_item in items:
            v = NvdSource.convert_vuln(cve_item)
            if v:
                data.append(v)
        return data

    @staticmethod
    def _convert_api(api_data: dict) -> list[Vulnerability]:
        """Convert NVD API data to vulnerability"""
        v = NvdSource.convert_api_vuln(api_data)
        return [v] if v else []

    def refresh(self):
        """Refresh CVE data"""
        return self.download_all()

    def bulk_search(self, app_info, pkg_list):
        """
        Bulk search the resource instead of downloading the information
        :return: Vulnerability result
        """

    @staticmethod
    def convert_vuln(vuln: dict) -> Vulnerability | None:
        vid = vuln["cve"]["CVE_data_meta"]["ID"]
        if filterable_cve(vid):
            return None
        assigner = vuln["cve"]["CVE_data_meta"]["ASSIGNER"]
        # CWE
        problem_type = ""
        if (
            vuln["cve"]["problemtype"]["problemtype_data"]
            and vuln["cve"]["problemtype"]["problemtype_data"][0]["description"]
        ):
            problem_type = vuln["cve"]["problemtype"]["problemtype_data"][0][
                "description"
            ][0]["value"]
        cvss_v3 = None
        severity = None
        base_score = None
        description = vuln["cve"]["description"]["description_data"][0]["value"]
        # Issue 12 - Ignore disputed vulnerabilities
        if "** DISPUTED **" in description:
            return None
        rdata = vuln.get("cve", {}).get("references", {}).get("reference_data", [])
        related_urls = [r["url"] for r in rdata]
        if "baseMetricV3" in vuln["impact"]:
            cvss_data = vuln["impact"]["baseMetricV3"]["cvssV3"]
            cvss_data["exploitabilityScore"] = vuln["impact"]["baseMetricV3"][
                "exploitabilityScore"
            ]
            cvss_data["impactScore"] = vuln["impact"]["baseMetricV3"]["impactScore"]
            cvss_v3 = CvssV3(
                base_score=cvss_data["baseScore"],
                exploitability_score=cvss_data["exploitabilityScore"],
                impact_score=cvss_data["impactScore"],
                attack_vector=cvss_data["attackVector"],
                attack_complexity=cvss_data["attackComplexity"],
                privileges_required=cvss_data["privilegesRequired"],
                user_interaction=cvss_data["userInteraction"],
                scope=cvss_data["scope"],
                confidentiality_impact=cvss_data["confidentialityImpact"],
                integrity_impact=cvss_data["integrityImpact"],
                availability_impact=cvss_data["availabilityImpact"],
                vector_string=cvss_data["vectorString"],
            )
            severity = cvss_data["baseSeverity"]
            base_score = cvss_v3.base_score
        details = NvdSource.convert_vuln_detail(vuln)
        if not details:
            return None
        return Vulnerability(
            vid,
            assigner,
            problem_type,
            base_score,
            severity,
            description,
            related_urls,
            details,
            cvss_v3,
            vuln["lastModifiedDate"],
            vuln["publishedDate"],
        )

    @staticmethod
    def convert_vuln_detail(vuln: dict) -> list[VulnerabilityDetail] | None:
        nodes_list = vuln["configurations"]["nodes"]
        details = []
        for node in nodes_list:
            cpe_list = []
            # For AND operator we store all the cpe_matches thus
            # increasing the false-positives. But this is better than leaving
            # the CPE out altogether. Grafeas format unfortunately is not
            # suitable for AND/OR based vulnerability storage
            # min and max_affected_version can sometimes include the excluded version
            # thus, further increasing the false positives
            if node["operator"] == "AND":
                for cc in node.get("children", []):
                    cpe_list += cc["cpe_match"]
            cpe_list += node.get("cpe_match", [])
            cpe_details_list = []
            fix_cpe_uri = None
            for cpe in cpe_list:
                detail = {}
                if not cpe.get("cpe23Uri"):
                    continue
                if cpe["vulnerable"]:
                    detail["cpe_uri"] = clean_cpe_uri(cpe.get("cpe23Uri"))
                    detail["mii"] = cpe.get("versionStartIncluding")
                    detail["mie"] = cpe.get("versionStartExcluding")
                    detail["mai"] = cpe.get("versionEndIncluding")
                    detail["mae"] = cpe.get("versionEndExcluding")
                    detail["source_update_time"] = vuln["lastModifiedDate"]
                    detail["source_orig_time"] = vuln["publishedDate"]
                    cpe_details_list.append(detail)
                else:  # cpe is not vulnerable
                    if node["operator"] == "OR":
                        fix_cpe_uri = clean_cpe_uri(cpe["cpe23Uri"])
            # Add fix version details
            for det in cpe_details_list:
                if fix_cpe_uri:
                    det["fixed_location"] = fix_cpe_uri
                adetail = VulnerabilityDetail.from_dict(det)
                if adetail:
                    details.append(adetail)
        if not details:
            return None
        return details

    @staticmethod
    def convert_api_vuln_detail(vuln: dict) -> list[VulnerabilityDetail] | None:
        """Convert configurations section from the API into details"""
        config_list = vuln.get("configurations", [])
        details = []
        rdata = vuln.get("references", [])
        git_urls = [r["url"].lower() for r in rdata if "git" in r["url"]]
        # Alternative CPEs identified for the given vulnerability detail
        vuln_alt_cpes = {}
        for aconfig in config_list:
            cpe_list = []
            nodes = aconfig.get("nodes", [])
            oper = aconfig.get("operator", "")
            # AND operator flow
            if oper == "AND":
                for cc in nodes:
                    if not cc.get("negate"):
                        cpe_list += filter(
                            lambda c: c.get("vulnerable"), cc["cpeMatch"]
                        )
            for anode in nodes:
                if not anode.get("negate"):
                    cpe_list += filter(
                        lambda c: c.get("vulnerable"), anode.get("cpeMatch", [])
                    )
            cpe_details_list = []
            for cpe in cpe_list:
                cpe_uri = cpe["criteria"]
                # Ignore os and hardware vulnerabilities from nvd
                if (
                    cpe_uri
                    and cpe_uri.startswith("cpe:2.3:o")
                    or cpe_uri.startswith("cpe:2.3:h")
                ):
                    continue
                all_parts = CPE_FULL_REGEX.match(cpe_uri)
                # If a single version is mentioned using cpe then use that as a fallback
                single_version = ""
                if (
                    all_parts
                    and all_parts.group("version")
                    and all_parts.group("version") != "*"
                ):
                    single_version = all_parts.group("version")
                    # Version numbers could have erroneous \ or commas
                    if single_version:
                        single_version = (
                            single_version.removeprefix(",").removesuffix("\\").strip()
                        )
                version_start_including = cpe.get(
                    "versionStartIncluding", single_version
                )
                version_end_including = cpe.get("versionEndIncluding", single_version)
                detail = {
                    "cpe_uri": clean_cpe_uri(cpe_uri),
                    "mii": version_start_including,
                    "mie": cpe.get("versionStartExcluding"),
                    "mai": version_end_including,
                    "mae": cpe.get("versionEndExcluding"),
                    "source_update_time": vuln["lastModified"],
                    "source_orig_time": vuln["published"],
                }
                cpe_details_list.append(detail)
                alt_cpes = get_alt_cpes(detail["cpe_uri"], git_urls)
                for altc in alt_cpes:
                    # Filter duplicates
                    if vuln_alt_cpes.get(altc):
                        continue
                    new_git_detail = detail.copy()
                    new_git_detail["cpe_uri"] = altc
                    cpe_details_list.append(new_git_detail)
                    vuln_alt_cpes[altc] = True
            for det in cpe_details_list:
                adetail = VulnerabilityDetail.from_dict(det)
                if adetail:
                    details.append(adetail)
        if not details:
            return None
        return details

    @staticmethod
    def _get_value(vuln: dict, key: str) -> str:
        if vuln.get(key):
            value = vuln[key]
            if isinstance(value, str):
                return value
            if isinstance(value, list):
                for val in value:
                    if isinstance(val, dict):
                        if val.get("lang", "") == "en":
                            return val.get("value")
        return ""

    @staticmethod
    def convert_api_vuln(vuln: dict) -> Vulnerability | None:
        # Undergoing Analysis, Analyzed, Modified, Deferred are accepted.
        # Only the below 3 can be rejected
        if vuln["vulnStatus"] in ("Awaiting Analysis", "Rejected", "Received"):
            return None
        vid = vuln["id"]
        assigner = vuln["sourceIdentifier"]
        # CWE
        problem_type = ""
        for aweakness in vuln.get("weaknesses", []):
            if aweakness.get("description") and aweakness["description"][0][
                "value"
            ].startswith("CWE-"):
                problem_type = aweakness["description"][0]["value"]
                break
        description = NvdSource._get_value(vuln, "descriptions")
        # Ignore disputed CVEs. Eg: CVE-2023-35116
        # CVE-2023-39017 uses the phrase "this is disputed"
        for ds in (
            "** DISPUTED **",
            "this is not a valid vulnerability report",
            "this is disputed",
        ):
            if ds in description:
                return None
        affected_symbols = extract_affected_symbols(description)
        rdata = vuln.get("references", [])
        related_urls = [r["url"] for r in rdata]
        metrics = vuln.get("metrics", {})
        cvss_metrics = []
        if metrics.get("cvssMetricV31"):
            cvss_metrics = metrics.get("cvssMetricV31")
        elif metrics.get("cvssMetricV30"):
            cvss_metrics = metrics.get("cvssMetricV30")
        if not cvss_metrics:
            return None
        cvss_data = cvss_metrics[0].get("cvssData")
        cvss_v3 = CvssV3(
            base_score=cvss_data["baseScore"],
            exploitability_score=cvss_metrics[0]["exploitabilityScore"],
            impact_score=cvss_metrics[0]["impactScore"],
            attack_vector=cvss_data["attackVector"],
            attack_complexity=cvss_data["attackComplexity"],
            privileges_required=cvss_data["privilegesRequired"],
            user_interaction=cvss_data["userInteraction"],
            scope=cvss_data["scope"],
            confidentiality_impact=cvss_data["confidentialityImpact"],
            integrity_impact=cvss_data["integrityImpact"],
            availability_impact=cvss_data["availabilityImpact"],
            vector_string=cvss_data["vectorString"],
        )
        severity = cvss_data["baseSeverity"]
        base_score = cvss_v3.base_score
        details = NvdSource.convert_api_vuln_detail(vuln)
        if not details:
            return None
        return Vulnerability(
            vid,
            assigner,
            problem_type,
            base_score,
            severity,
            description,
            related_urls,
            details,
            cvss_v3,
            vuln["lastModified"],
            vuln["published"],
            affected_symbols,
        )



import re
from typing import Any, Generator, Dict, List, Tuple

import apsw
import orjson

from vdb.lib import db6, utils
from vdb.lib.config import PLACEHOLDER_EXCLUDE_VERSION, PLACEHOLDER_FIX_VERSION
from vdb.lib.cve_model import CVE, CVE1, Product, Versions, Status
from vdb.lib.utils import load_json, vers_compare


IS_ADVISORY = re.compile("^[A-Z]{1,7}-")
BATCH_SIZE = 200

def _process_hits(raw_hits: List[Tuple]) -> List[Dict[str, Any]]:
    filtered_list = []
    for ahit in raw_hits:
        cve_id = ahit[0]
        vers = ahit[4]
        filtered_list.append(
            {
                "cve_id": cve_id,
                "type": ahit[1],
                "namespace": ahit[2],
                "name": ahit[3],
                "vers": vers,
                "purl_prefix": ahit[-1],
            }
        )
    return filtered_list


def get_unaffected(source_data, vers):
    if source_data and source_data.root.containers:
        products: List[Product] = source_data.root.containers.cna.affected.root
        for p in products:
            versions: List[Versions] = p.versions
            if versions:
                for ver in versions:
                    if ver.status == Status.unaffected:
                        if ver.version:
                            return ver.version.root
    if "|" in vers:
        vers = vers.split("|")[-1]
        if "!=" in vers or "<=" not in vers:
            return vers.replace("<", "").replace("!=", "")
    elif "/<" in vers and "/<=" not in vers:
        return vers.split("/<")[-1]
    elif vers.endswith(f"<{PLACEHOLDER_EXCLUDE_VERSION}"):
        return PLACEHOLDER_EXCLUDE_VERSION
    elif vers.endswith(f"<{PLACEHOLDER_FIX_VERSION}"):
        return PLACEHOLDER_FIX_VERSION
    return ""


def get_cve_data(
    db_conn: apsw.Connection | None, index_hits: List, search_str: str
) -> Generator:
    """Get CVE data for the index results

    Args:
        db_conn: DB Connection or None to create a new one
        index_hits: Hits from one of the search methods
        search_str: Original search string used

    Returns:
        generator: generator for CVE data with original source data as a pydantic model
    """
    if not db_conn:
        db_conn, _ = db6.get(read_only=True)
    for ahit in index_hits:
        results = exec_query(
            db_conn,
            "SELECT DISTINCT cve_id, type, namespace, name, source_data_hash, source_data, override_data, vers, purl_prefix FROM cve_data WHERE cve_id = ? AND vers = ? AND purl_prefix = ? GROUP BY purl_prefix ORDER BY cve_id DESC",
            (ahit["cve_id"], ahit["vers"], ahit["purl_prefix"]),
        )
        for res in results:
            source_data = (
                CVE(root=CVE1.model_validate(apsw.jsonb_decode(res[5]), strict=False))
                if res[5]
                else None
            )
            vers = res[7]
            fix_version = get_unaffected(source_data, vers)
            yield {
                "cve_id": res[0],
                "type": res[1],
                "namespace": res[2],
                "name": res[3],
                "matching_vers": ahit["vers"],
                "matched_by": search_str,
                "source_data_hash": res[4],
                "source_data": source_data,
                "override_data": (apsw.jsonb_decode(res[6]) if res[6] else None),
                "vers": vers,
                "purl_prefix": res[8],
                "fix_version": fix_version,
            }


def get_cve_data_batched(
    db_conn: apsw.Connection | None, index_hits: List, search_str: str
) -> Generator:
    """
    Get CVE data for the index results efficiently.
    Uses batching and CTEs to avoid N+1 query performance issues.
    """
    if not index_hits:
        return

    if not db_conn:
        db_conn, _ = db6.get(read_only=True)
    total_hits = len(index_hits)
    for i in range(0, total_hits, BATCH_SIZE):
        batch = index_hits[i : i + BATCH_SIZE]
        query_args = []
        for hit in batch:
            query_args.extend([hit["cve_id"], hit["vers"], hit["purl_prefix"]])
        placeholders = ",".join(["(?,?,?)"] * len(batch))
        sql = f"""
            WITH lookup(cve_id, vers, purl_prefix) AS (
                VALUES {placeholders}
            )
            SELECT
                d.cve_id, 
                d.type, 
                d.namespace, 
                d.name, 
                d.source_data_hash, 
                d.source_data, 
                d.override_data, 
                d.vers, 
                d.purl_prefix 
            FROM cve_data d
            JOIN lookup l 
                ON d.cve_id = l.cve_id 
                AND d.vers = l.vers 
                AND d.purl_prefix = l.purl_prefix
            GROUP BY d.cve_id, d.vers, d.purl_prefix
            ORDER BY d.cve_id DESC;
        """
        results = exec_query(db_conn, sql, tuple(query_args))
        for res in results:
            source_data_str = res[5]
            override_data_str = res[6]
            source_data = (
                CVE(root=CVE1.model_validate(apsw.jsonb_decode(source_data_str), strict=False))
                if source_data_str
                else None
            )
            vers = res[7]
            fix_version = get_unaffected(source_data, vers)
            yield {
                "cve_id": res[0],
                "type": res[1],
                "namespace": res[2],
                "name": res[3],
                "matching_vers": vers,
                "matched_by": search_str,
                "source_data_hash": res[4],
                "source_data": source_data,
                "override_data": (apsw.jsonb_decode(override_data_str) if override_data_str else None),
                "vers": vers,
                "purl_prefix": res[8],
                "fix_version": fix_version,
            }


def search_by_any(any_str: str, with_data: bool = False) -> List:
    """Convenient method to search by a string"""
    if any_str.startswith("pkg:"):
        return search_by_purl_like(any_str, with_data)
    if IS_ADVISORY.search(any_str):
        return search_by_cve(any_str, with_data)
    if any_str.startswith("http"):
        return search_by_url(any_str, with_data)
    return search_by_cpe_like(any_str, with_data)


def search_by_cpe_like(cpe: str, with_data: bool = False) -> List:
    """Search by CPE or colon-separate strings"""
    db_conn, index_conn = db6.get(read_only=True)
    if cpe.startswith("cpe:"):
        vendor, package, version, _ = utils.parse_cpe(cpe)
    elif cpe.count(":") == 2:
        vendor, package, version = cpe.split(":")
    else:
        return []
    # check for vendor name in both namespace and type
    raw_hits = exec_query(
        index_conn,
        "SELECT DISTINCT cve_id, type, namespace, name, vers, purl_prefix FROM cve_index where (namespace = ? OR type = ?) AND name = ? AND vers_compare(?, vers)",
        (vendor, vendor, package, version),
    )
    filtered_list = _process_hits(raw_hits)
    if with_data:
        return list(get_cve_data_batched(db_conn, filtered_list, cpe))
    return filtered_list


def search_by_purl_like(purl: str, with_data: bool = False) -> List:
    """Search by purl like string"""
    db_conn, index_conn = db6.get(read_only=True)
    purl_obj = utils.parse_purl(purl)
    if purl_obj:
        ptype = purl_obj.get("type")
        namespace = purl_obj.get("namespace")
        name = purl_obj.get("name")
        version = purl_obj.get("version", "*")
        purl_prefix = f"pkg:{ptype}/"
        if namespace:
            purl_prefix = f'{purl_prefix}{namespace.replace("@", "%40")}/'
        # Handle distro names for linux os purls by prefixing distro name to name
        if purl_obj.get("qualifiers") and purl_obj["qualifiers"].get("distro_name"):
            distro_name = purl_obj["qualifiers"].get("distro_name")
            name = f"{distro_name}/{name}"
        purl_prefix = f"{purl_prefix}{name}"
        args = (purl_prefix, version,)
        raw_hits = exec_query(
            index_conn,
            "SELECT DISTINCT cve_id, type, namespace, name, vers, purl_prefix FROM cve_index where purl_prefix = ? AND vers_compare(?, vers)",
            args,
        )
        filtered_list = _process_hits(raw_hits)
        if with_data:
            return list(get_cve_data_batched(db_conn, filtered_list, purl))
        return filtered_list
    return []


def search_by_cve(
    cve_id: str, with_data: bool = False, with_limit: int | None = None
) -> List:
    """Search by CVE"""
    db_conn, index_conn = db6.get(read_only=True)
    filter_part = "cve_id LIKE ?" if "%" in cve_id else "cve_id = ?"
    filter_part = f"{filter_part} ORDER BY cve_id DESC"
    args = [cve_id]
    if with_limit and isinstance(with_limit, int):
        filter_part = f"{filter_part} LIMIT ?"
        args.append(with_limit)
    args = tuple(args)
    raw_hits = exec_query(
        index_conn,
        f"SELECT DISTINCT cve_id, type, namespace, name, vers, purl_prefix FROM cve_index where {filter_part}",
        args,
    )
    filtered_list = _process_hits(raw_hits)
    if with_data:
        return list(get_cve_data_batched(db_conn, filtered_list, cve_id))
    return filtered_list


def search_by_url(url: str, with_data: bool = False) -> List:
    """Search by URL"""
    purl_obj = utils.url_to_purl(url)
    if not purl_obj:
        return []
    name = purl_obj["name"]
    purl_str = (
        f"""pkg:{purl_obj['type']}/{purl_obj['namespace'].replace("@", "%40")}/{name}"""
        if purl_obj["namespace"]
        else f"pkg:{purl_obj['type']}/{name}"
    )
    if purl_obj["version"]:
        purl_str = f"{purl_str}@{purl_obj['version']}"
    return search_by_purl_like(purl_str, with_data)


def search_by_cdx_bom(bom_file: str, with_data: bool = False) -> Generator:
    """Search by CycloneDX BOM file"""
    cdx_obj = load_json(bom_file)
    for component in cdx_obj.get("components", []):
        if component.get("purl"):
            yield search_by_purl_like(component["purl"], with_data)
        if component.get("cpe"):
            yield search_by_cpe_like(component["cpe"], with_data)


def latest_malware(with_limit=20, with_data=False) -> Generator:
    """Search for latest malware with CVE ID beginning with MAL-"""
    yield search_by_cve("MAL-%", with_data=with_data, with_limit=with_limit)


def exec_query(conn, query: str, args: Tuple[str, ...]) -> list:
    res = conn.execute(query, args)
    return res.fetchall()



from vdb.lib.cve_model import Reference

REFERENCE_TAGS_MAP = {
    "exploit": [
        "exploit-db.com/",
        "exploit-database",
        "seebug.org",
        "seclists.org",
        "nu11secur1ty",
        "packetstormsecurity.com",
        "coresecurity.com",
        "project-zero",
        "0dd.zone",
        "snyk.io/research/",
        "chromium.googlesource.com/infra",
        "synacktiv.com",
        "bishopfox.com",
        "zerodayinitiative.com",
        "www.samba.org/samba/security/",
        "www.synology.com/support/security/",
        "us-cert.gov/advisories",
    ],
    "government-resource": [
        ".gov",
        "cisa",
        "kevc",
    ],
    "issue-tracking": [
        "bugzilla",
        "bugs.",
        "chat.",
        "/issues",
        "/merge_request",
        "oss-fuzz",
        "trac.",
        "security-tracker.",
        "/bugs",
    ],
    "mailing-list": [
        "openwall.com",
        "oss-security",
        "www.mail-archive.com",
        "lists.",
        "mail.",
        "/discussion/",
        "/archives/",
        "groups.",
        "/community",
        "/forum",
        "/discuss",
        "-announce",
    ],
    "mitigation": [],
    "not-applicable": [],
    "patch": [
        "/commit",
        "/pull",
        "/code/ci",
        ".patch",
    ],
    "media-coverage": [
        "blog",
        "support",
        "media",
        "tech-updates",
        "/news",
        "/article",
        ".html",
        "/entry",
        ".txt",
        "/comments/",
        "youtube.com",
        "medium.com",
        "twitter.com",
    ],
    "release-notes": ["/release", ".md", "/changeset", "releases/"],
    "technical-description": [
        "poc",
        "hackerone",
        "bugcrowd",
        "bounty",
        "huntr.dev",
        "bounties",
        "attackerkb",
        "support.",
        ".pdf",
        "docs.google.com",
    ],
    "third-party-advisory": [
        "research",
        "xss",
        "csrf",
        "ssrf",
        "sqli",
        "disclosure",
        "rapid7",
        "reference",
        ".me/",
    ],
    "vendor-advisory": [
        "oracle.com",
        "curl.haxx.se",
        "nodejs.org",
        "/security.",
        "/securityadvisories.",
        "sec-consult.com",
        "jenkins.io/security",
        "support.f5.com",
        "suricata-ids.org/",
        "foxitsoftware.com/support/",
        "success.trendmicro.com/",
        "docs.jamf.com/",
        "www.postgresql.org/about",
        "access.redhat.com",
        "support.apple.com",
        "rubyonrails-security",
        "usn.ubuntu.com",
        "security.gentoo.org",
        "debian.org",
        "apache.org",
        "gitlab.alpinelinux.org",
        "bugs.busybox.net",
        "/security-advisor",
        "/alert",
        "wordpress",
        "wpvulndb",
        "/bug/view/",
    ],
    "vdb-entry": [
        "/advisories",
        "/vulnerabilit",
        "cve-",
        "ghsa-",
        "dsa-",
        "mal-",
        "/vuln/",
        "portal.msrc.microsoft.com",
        "/id/",
        "/bid/",
        "kb.",
        "jvn.jp/",
        "vulndb",
        "vulncheck",
        "glsa",
        "rhsa-",
    ],
}


def get_reference_tags(ref_urls: list[str | Reference]) -> list[dict[str, str]]:
    """Tag the urls under references"""
    tags = []
    for aref in ref_urls:
        theurl = aref.url if isinstance(aref, Reference) else aref
        for reference_tag, tag_patterns in REFERENCE_TAGS_MAP.items():
            for keyword_str in tag_patterns:
                if keyword_str in theurl:
                    tags.append({"url": theurl, "tag": reference_tag})
                    break
    return tags



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
from datetime import datetime, timezone
import json
import logging
import os
import shutil
import sys
import types

from rich.console import Console
from rich.live import Live
from rich.table import Table

from vdb.lib import config, display, db6 as db_lib, search
from vdb.lib.aqua import AquaSource
from vdb.lib.gha import GitHubSource
from vdb.lib.osv import OSVSource

if sys.platform == "win32" and os.environ.get("PYTHONIOENCODING") is None:
    sys.stdin.reconfigure(encoding="utf-8")
    sys.stdout.reconfigure(encoding="utf-8")
    sys.stderr.reconfigure(encoding="utf-8")

ORAS_AVAILABLE = False
# oras is an optional dependency
try:
    from vdb.lib.orasclient import download_image

    ORAS_AVAILABLE = True
except ImportError:
    pass

console = Console()

logging.basicConfig(
    level=logging.INFO, format="%(levelname)s [%(asctime)s] %(message)s"
)
LOG = logging.getLogger(__name__)
for _ in ("httpx",):
    logging.getLogger(_).disabled = True

AT_LOGO = r"""
             ___
  /\  ._  ._  | |_  ._ _   _. _|_
 /--\ |_) |_) | | | | (/_ (_|  |_
      |   |
"""


def build_args():
    """
    Constructs command line arguments for the vulnerability-db tool
    """
    parser = argparse.ArgumentParser(
        description="AppThreat's vulnerability database and package search library with a sqlite storage."
    )
    parser.add_argument(
        "--clean",
        action="store_true",
        default=False,
        dest="clean",
        help="Clear the vulnerability database cache from platform specific user_data_dir.",
    )
    parser.add_argument(
        "--cache",
        action="store_true",
        default=False,
        dest="cache",
        help="Cache vulnerability information in platform specific user_data_dir.",
    )
    parser.add_argument(
        "--cache-os",
        action="store_true",
        default=False,
        dest="cache_os",
        help="Cache OS vulnerability information in platform specific user_data_dir.",
    )
    parser.add_argument(
        "--only-osv",
        action="store_true",
        default=False,
        dest="only_osv",
        help="Use only OSV as the source. Use with --cache.",
    )
    parser.add_argument(
        "--only-aqua",
        action="store_true",
        default=False,
        dest="only_aqua",
        help="Use only Aqua vuln-list as the source. Use with --cache.",
    )
    parser.add_argument(
        "--only-ghsa",
        action="store_true",
        default=False,
        dest="only_ghsa",
        help="Use only recent ghsa as the source. Use with --cache.",
    )
    parser.add_argument(
        "--search",
        dest="search",
        help="Search for the package or vulnerability ID (CVE, GHSA, ALSA, DSA, etc.) in the database. Use purl, cpe, or git http url.",
    )
    parser.add_argument(
        "--list-malware",
        action="store_true",
        default=False,
        dest="list_malware",
        help="List latest malwares with CVE ID beginning with MAL-.",
    )
    parser.add_argument(
        "--bom",
        dest="bom_file",
        help="Search for packages in the CycloneDX BOM file.",
    )
    parser.add_argument(
        "--download-image",
        action="store_true",
        default=False,
        dest="download_image",
        help="Downloaded pre-created vdb image to platform specific user_data_dir. Application vulnerabilities only.",
    )
    parser.add_argument(
        "--download-full-image",
        action="store_true",
        default=False,
        dest="download_full_image",
        help="Downloaded pre-created vdb image to platform specific user_data_dir. All vulnerabilities including OS.",
    )
    parser.add_argument(
        "--print-vdb-metadata",
        action="store_true",
        default=False,
        dest="print_vdb_metadata",
        help="Display metadata about the current vdb in user_data_dir.",
    )
    return parser.parse_args()


def print_results(results):
    added_row_keys = {}
    table = Table(title="VDB Results", show_lines=True)
    table.add_column("CVE", justify="left", max_width=20)
    table.add_column("Locators")
    table.add_column("Fix Version")
    table.add_column("Description")
    table.add_column("Affected Symbols", max_width=50)
    if isinstance(results, types.GeneratorType):
        with Live(
            table, console=console, refresh_per_second=4, vertical_overflow="visible"
        ):
            for result_gen in results:
                if isinstance(result_gen, dict):
                    display.add_table_row(table, result_gen, added_row_keys)
                elif isinstance(result_gen, list):
                    for res in result_gen:
                        display.add_table_row(table, res, added_row_keys)
                elif isinstance(result_gen, types.GeneratorType):
                    for res in result_gen:
                        display.add_table_row(table, res, added_row_keys)
    elif isinstance(results, list):
        for res in results:
            display.add_table_row(table, res, added_row_keys)
        console.print(table)


def create_db_file_metadata(sources, cve_data_count, cve_index_count):
    """Method to create the vdb file metadata"""
    metadata = {
        "created_utc": datetime.now(tz=timezone.utc).isoformat(timespec="seconds"),
        "cve_data_count": cve_data_count,
        "cve_index_count": cve_index_count,
        "sources": [s.__class__.__name__ for s in sources],
    }
    include_list = []
    ignore_list = []
    # Collect injected metadata from environment variables
    for name, value in os.environ.items():
        if value is None:
            continue
        if name.startswith("VDB_METADATA_"):
            if value in ("true", "1"):
                value = True
            elif value in ("false", "0"):
                value = False
            metadata[name.replace("VDB_METADATA_", "").lower()] = value
        elif name.startswith("VDB_IGNORE_") and value in ("true", "1"):
            ignore_list.append(name.replace("VDB_IGNORE_", "").lower())
        elif name.startswith("VDB_EXCLUDE_") and value in ("true", "1"):
            ignore_list.append(name.replace("VDB_EXCLUDE_", "").lower())
        elif name.startswith("VDB_INCLUDE_") and value in ("true", "1"):
            include_list.append(name.replace("VDB_INCLUDE_", "").lower())
    if include_list:
        metadata["include_list"] = include_list
    if ignore_list:
        metadata["ignore_list"] = ignore_list
    if os.getenv("NVD_START_YEAR"):
        metadata["start_year"] = os.getenv("NVD_START_YEAR")
    if os.getenv("GITHUB_PAGE_COUNT"):
        metadata["github_page_count"] = os.getenv("GITHUB_PAGE_COUNT")
    return metadata


def print_db_file_metadata(metadata_file):
    if not os.path.exists(metadata_file):
        return
    with open(metadata_file, encoding="utf-8") as fp:
        db_meta = json.load(fp)
        table = Table(
            title="VDB Summary",
            show_lines=True,
            caption=f"Metadata file: {metadata_file}",
        )
        table.add_column("Property")
        table.add_column("Value")
        for k, v in db_meta.items():
            table.add_row(k, str(v))
        console.print(table)


def main():
    """Main function"""
    args = build_args()
    print(AT_LOGO)
    if args.clean:
        if os.path.exists(config.DATA_DIR):
            shutil.rmtree(config.DATA_DIR, ignore_errors=True)
        if not os.path.exists(config.DATA_DIR):
            LOG.info("VDB cache cleaned successfully.")
        else:
            LOG.info("VDB cache at %s not cleaned successfully.", config.DATA_DIR)
    if args.print_vdb_metadata:
        print_db_file_metadata(config.VDB_METADATA_FILE)
    if args.download_image or args.download_full_image:
        db_url = (
            config.VDB_DATABASE_URL
            if args.download_full_image
            else config.VDB_APP_ONLY_DATABASE_URL
        )
        if ORAS_AVAILABLE:
            LOG.info(
                "Downloading vdb image from %s to %s",
                db_url,
                config.DATA_DIR,
            )
            try:
                download_image(db_url, config.DATA_DIR)
                print_db_file_metadata(config.VDB_METADATA_FILE)
            except Exception as e:
                LOG.info(
                    "Vulnerability database download using oras failed. Please try again."
                )
                LOG.exception(e)
                return
        else:
            console.print(
                "Oras library is not available. Install using 'pip install appthreat-vulnerability-db[oras]' and re-run this command.",
                markup=False,
            )
    elif args.cache or args.cache_os:
        db_lib.get()
        db_lib.clear_all()
        if args.only_osv:
            sources = [OSVSource()]
        elif args.only_aqua:
            sources = [AquaSource()]
        elif args.only_ghsa:
            sources = [GitHubSource()]
        else:
            sources = [OSVSource(), GitHubSource()]
        # AquaSource also includes NVD
        if args.cache_os and not args.only_aqua:
            sources.insert(0, AquaSource())
        for s in sources:
            LOG.info("Refreshing %s", s.__class__.__name__)
            s.refresh()
        cve_data_count, cve_index_count = db_lib.stats()
        console.print(
            "cve_data_count", cve_data_count, "cve_index_count", cve_index_count
        )
        db_lib.optimize_and_close_all()
        # Create metadata about the database
        with open(config.VDB_METADATA_FILE, mode="w", encoding="utf-8") as meta_file:
            json.dump(
                create_db_file_metadata(sources, cve_data_count, cve_index_count),
                meta_file,
            )
    if args.search:
        if db_lib.needs_update():
            console.print(
                "Vulnerability database needs to be refreshed. Please run 'vdb --download-image' to download the latest."
            )
        results = search.search_by_any(args.search, with_data=True)
        if results:
            print_results(results)
        else:
            console.print("No results found!")
    elif args.bom_file:
        if os.path.exists(args.bom_file):
            results_generator = search.search_by_cdx_bom(args.bom_file, with_data=True)
            print_results(results_generator)
    elif args.list_malware:
        results_generator = search.latest_malware(with_data=True)
        print_results(results_generator)


if __name__ == "__main__":
    main()



from datetime import datetime, timedelta, timezone
import json
import os
import sys
import tempfile
import apsw

from vdb.lib import config
from vdb.lib.utils import vers_compare

db_conn: apsw.Connection = None
index_conn: apsw.Connection = None
tables_created = False
DB_FILE_SEP = "///" if sys.platform == "win32" else "//"


def ensure_schemas(db_conn_obj: apsw.Connection, index_conn_obj: apsw.Connection):
    # Use temp files to reduce memory usage during VACCUM
    # See issue #203
    temp_store_dir = os.getenv("VDB_TEMP_DIR", tempfile.mkdtemp(prefix="vdb-temp"))
    """Create the sqlite tables and indexes in case they don't exist"""
    db_conn_obj.execute(
        "CREATE TABLE if not exists cve_data(cve_id TEXT NOT NULL, type TEXT NOT NULL, namespace TEXT, name TEXT NOT NULL, source_data BLOB NOT NULL, override_data BLOB, source_data_hash TEXT NOT NULL, vers TEXT NOT NULL, purl_prefix TEXT NOT NULL);"
    )
    db_conn_obj.pragma("synchronous", "OFF")
    db_conn_obj.pragma("journal_mode", "MEMORY")
    if os.access(temp_store_dir, os.W_OK):
        db_conn_obj.pragma("temp_store", "FILE")
        db_conn_obj.pragma("temp_store_directory", temp_store_dir)
    else:
        db_conn_obj.pragma("temp_store", "MEMORY")
    index_conn_obj.execute(
        "CREATE TABLE if not exists cve_index(cve_id TEXT NOT NULL, type TEXT NOT NULL, namespace TEXT, name TEXT NOT NULL, vers TEXT NOT NULL, purl_prefix TEXT NOT NULL);"
    )
    index_conn_obj.pragma("synchronous", "OFF")
    index_conn_obj.pragma("journal_mode", "MEMORY")
    if os.access(temp_store_dir, os.W_OK):
        index_conn_obj.pragma("temp_store", "FILE")
        index_conn_obj.pragma("temp_store_directory", temp_store_dir)
    else:
        index_conn_obj.pragma("temp_store", "MEMORY")

def get(
    db_file: str = config.VDB_BIN_FILE,
    index_file: str = config.VDB_BIN_INDEX,
    read_only=False,
) -> (apsw.Connection, apsw.Connection):
    """Gets the connection to the index and the data databases. Raises apsw.CantOpenError if the database is not available."""
    global db_conn, index_conn, tables_created
    if not db_file.startswith("file:") and db_file != ":memory:":
        db_file = f"file:{DB_FILE_SEP}{os.path.abspath(db_file)}"
    if not index_file.startswith("file:") and index_file != ":memory:":
        index_file = f"file:{DB_FILE_SEP}{os.path.abspath(index_file)}"
    rw_flags = (
        apsw.SQLITE_OPEN_URI
        | apsw.SQLITE_OPEN_NOFOLLOW
        | apsw.SQLITE_OPEN_CREATE
        | apsw.SQLITE_OPEN_READWRITE
    )
    ro_flags = (
        apsw.SQLITE_OPEN_URI | apsw.SQLITE_OPEN_NOFOLLOW | apsw.SQLITE_OPEN_READONLY
    )
    # To avoid apsw.CantOpenError, we support readonly mode only if the db exists
    flags = (
        ro_flags
        if read_only and os.path.exists(db_file) and os.path.exists(index_file)
        else rw_flags
    )
    if not db_conn:
        db_conn = apsw.Connection(db_file, flags=flags)
    if not index_conn:
        index_conn = apsw.Connection(index_file, flags=flags)
        index_conn.createscalarfunction("vers_compare", vers_compare, deterministic=True)
    if not tables_created:
        ensure_schemas(db_conn, index_conn)
        tables_created = True
    return db_conn, index_conn


def stats():
    cve_data_count = 0
    res = db_conn.execute("SELECT count(*) FROM cve_data").fetchone()
    if res:
        cve_data_count = res[0]
    cve_index_count = 0
    res = index_conn.execute("SELECT count(*) FROM cve_index").fetchone()
    if res:
        cve_index_count = res[0]
    return cve_data_count, cve_index_count


def clear_all():
    if db_conn:
        db_conn.execute("DELETE FROM cve_data;")
    if index_conn:
        index_conn.execute("DELETE FROM cve_index;")


def optimize_and_close_all():
    """
    Safely close the connections by creating indexes and vacuuming if needed.
    """
    if db_conn:
        db_conn.execute(
            "CREATE INDEX if not exists idx1 on cve_data(cve_id, vers, purl_prefix);"
        )
        db_conn.execute("VACUUM;")
        db_conn.close()
    if index_conn:
        index_conn.execute("CREATE INDEX if not exists cidx1 on cve_index(cve_id);")
        index_conn.execute(
            "CREATE INDEX if not exists cidx2 on cve_index(type, namespace, name);"
        )
        index_conn.execute("CREATE INDEX if not exists cidx3 on cve_index(type, name);")
        index_conn.execute(
            "CREATE INDEX if not exists cidx4 on cve_index(namespace, name);"
        )
        index_conn.execute(
            "CREATE INDEX if not exists cidx5 on cve_index(purl_prefix);"
        )
        index_conn.execute("VACUUM;")
        index_conn.close()


def get_db_file_metadata():
    """
    Returns the metadata for the cached vdb files.
    """
    if not os.path.exists(config.VDB_METADATA_FILE):
        return None
    with open(config.VDB_METADATA_FILE, encoding="utf-8") as meta_file:
        return json.load(meta_file)


def needs_update(days=1, hours=0, minutes=0, seconds=0, default_status=False):
    """
    Determines if the db needs to be updated based on creation time
    """
    db_metadata = get_db_file_metadata()
    if not db_metadata or not db_metadata.get("created_utc"):
        return default_status
    current_time = datetime.now(tz=timezone.utc)
    created_time = datetime.fromisoformat(db_metadata["created_utc"])
    diff = current_time - created_time
    return diff > timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)



# pylint: disable=C0115, C0103, C0301

from __future__ import annotations

from enum import Enum
from typing import Annotated, Optional

from pydantic import BaseModel, Field

from vdb.lib.cve_model.common import (
    AttackComplexityType,
    AttackRequirementsType,
    AttackVectorType,
    AutomatableType,
    CiaRequirementType,
    CriticalScoreType,
    CriticalSeverityType,
    ExploitMaturityType,
    HighScoreType,
    HighSeverityType,
    LowScoreType,
    LowSeverityType,
    MediumScoreType,
    MediumSeverityType,
    ModifiedAttackComplexityType,
    ModifiedAttackRequirementsType,
    ModifiedAttackVectorType,
    ModifiedPrivilegesRequiredType,
    ModifiedSubCType,
    ModifiedSubIaType,
    ModifiedUserInteractionType,
    ModifiedVulnCiaType,
    NoneScoreType,
    NoneSeverityType,
    PrivilegesRequiredType,
    ProviderUrgencyType,
    RecoveryType,
    SafetyType,
    ScoreType,
    SeverityType,
    SubCiaType,
    UserInteractionType,
    ValueDensityType,
    VulnCiaType,
    VulnerabilityResponseEffortType,
)


class Version(Enum):
    field_4_0 = "4.0"


class Field0(BaseModel):
    environmentalScore: Optional[NoneScoreType] = None
    environmentalSeverity: Optional[NoneSeverityType] = None


class Field0Model(BaseModel):
    environmentalScore: Optional[LowScoreType] = None
    environmentalSeverity: Optional[LowSeverityType] = None


class Field0Model1(BaseModel):
    environmentalScore: Optional[MediumScoreType] = None
    environmentalSeverity: Optional[MediumSeverityType] = None


class Field0Model2(BaseModel):
    environmentalScore: Optional[HighScoreType] = None
    environmentalSeverity: Optional[HighSeverityType] = None


class Field0Model3(BaseModel):
    environmentalScore: Optional[CriticalScoreType] = None
    environmentalSeverity: Optional[CriticalSeverityType] = None


class Field0Model4(BaseModel):
    version: Annotated[Version, Field(description="CVSS Version")]
    vectorString: Annotated[
        str,
        Field(
            pattern="^CVSS:4[.]0/AV:[NALP]/AC:[LH]/AT:[NP]/PR:[NLH]/UI:[NPA]/VC:[HLN]/VI:[HLN]/VA:[HLN]/SC:[HLN]/SI:[HLN]/SA:[HLN](/E:[XAPU])?(/CR:[XHML])?(/IR:[XHML])?(/AR:[XHML])?(/MAV:[XNALP])?(/MAC:[XLH])?(/MAT:[XNP])?(/MPR:[XNLH])?(/MUI:[XNPA])?(/MVC:[XNLH])?(/MVI:[XNLH])?(/MVA:[XNLH])?(/MSC:[XNLH])?(/MSI:[XNLHS])?(/MSA:[XNLHS])?(/S:[XNP])?(/AU:[XNY])?(/R:[XAUI])?(/V:[XDC])?(/RE:[XLMH])?(/U:(X|Clear|Green|Amber|Red))?$"
        ),
    ]
    baseScore: ScoreType
    baseSeverity: SeverityType
    attackVector: Optional[AttackVectorType] = None
    attackComplexity: Optional[AttackComplexityType] = None
    attackRequirements: Optional[AttackRequirementsType] = None
    privilegesRequired: Optional[PrivilegesRequiredType] = None
    userInteraction: Optional[UserInteractionType] = None
    vulnConfidentialityImpact: Optional[VulnCiaType] = None
    vulnIntegrityImpact: Optional[VulnCiaType] = None
    vulnAvailabilityImpact: Optional[VulnCiaType] = None
    subConfidentialityImpact: Optional[SubCiaType] = None
    subIntegrityImpact: Optional[SubCiaType] = None
    subAvailabilityImpact: Optional[SubCiaType] = None
    exploitMaturity: Optional[ExploitMaturityType] = "NOT_DEFINED"
    confidentialityRequirement: Optional[CiaRequirementType] = "NOT_DEFINED"
    integrityRequirement: Optional[CiaRequirementType] = "NOT_DEFINED"
    availabilityRequirement: Optional[CiaRequirementType] = "NOT_DEFINED"
    modifiedAttackVector: Optional[ModifiedAttackVectorType] = "NOT_DEFINED"
    modifiedAttackComplexity: Optional[ModifiedAttackComplexityType] = "NOT_DEFINED"
    modifiedAttackRequirements: Optional[ModifiedAttackRequirementsType] = "NOT_DEFINED"
    modifiedPrivilegesRequired: Optional[ModifiedPrivilegesRequiredType] = "NOT_DEFINED"
    modifiedUserInteraction: Optional[ModifiedUserInteractionType] = "NOT_DEFINED"
    modifiedVulnConfidentialityImpact: Optional[ModifiedVulnCiaType] = "NOT_DEFINED"
    modifiedVulnIntegrityImpact: Optional[ModifiedVulnCiaType] = "NOT_DEFINED"
    modifiedVulnAvailabilityImpact: Optional[ModifiedVulnCiaType] = "NOT_DEFINED"
    modifiedSubConfidentialityImpact: Optional[ModifiedSubCType] = "NOT_DEFINED"
    modifiedSubIntegrityImpact: Optional[ModifiedSubIaType] = "NOT_DEFINED"
    modifiedSubAvailabilityImpact: Optional[ModifiedSubIaType] = "NOT_DEFINED"
    Safety: Optional[SafetyType] = "NOT_DEFINED"
    Automatable: Optional[AutomatableType] = "NOT_DEFINED"
    Recovery: Optional[RecoveryType] = "NOT_DEFINED"
    valueDensity: Optional[ValueDensityType] = "NOT_DEFINED"
    vulnerabilityResponseEffort: Optional[VulnerabilityResponseEffortType] = (
        "NOT_DEFINED"
    )
    providerUrgency: Optional[ProviderUrgencyType] = "NOT_DEFINED"


class Field0Model5(Field0, Field0Model4):
    version: Annotated[Version, Field(description="CVSS Version")]
    vectorString: Annotated[
        str,
        Field(
            pattern="^CVSS:4[.]0/AV:[NALP]/AC:[LH]/AT:[NP]/PR:[NLH]/UI:[NPA]/VC:[HLN]/VI:[HLN]/VA:[HLN]/SC:[HLN]/SI:[HLN]/SA:[HLN](/E:[XAPU])?(/CR:[XHML])?(/IR:[XHML])?(/AR:[XHML])?(/MAV:[XNALP])?(/MAC:[XLH])?(/MAT:[XNP])?(/MPR:[XNLH])?(/MUI:[XNPA])?(/MVC:[XNLH])?(/MVI:[XNLH])?(/MVA:[XNLH])?(/MSC:[XNLH])?(/MSI:[XNLHS])?(/MSA:[XNLHS])?(/S:[XNP])?(/AU:[XNY])?(/R:[XAUI])?(/V:[XDC])?(/RE:[XLMH])?(/U:(X|Clear|Green|Amber|Red))?$"
        ),
    ]
    baseScore: ScoreType
    baseSeverity: SeverityType
    attackVector: Optional[AttackVectorType] = None
    attackComplexity: Optional[AttackComplexityType] = None
    attackRequirements: Optional[AttackRequirementsType] = None
    privilegesRequired: Optional[PrivilegesRequiredType] = None
    userInteraction: Optional[UserInteractionType] = None
    vulnConfidentialityImpact: Optional[VulnCiaType] = None
    vulnIntegrityImpact: Optional[VulnCiaType] = None
    vulnAvailabilityImpact: Optional[VulnCiaType] = None
    subConfidentialityImpact: Optional[SubCiaType] = None
    subIntegrityImpact: Optional[SubCiaType] = None
    subAvailabilityImpact: Optional[SubCiaType] = None
    exploitMaturity: Optional[ExploitMaturityType] = "NOT_DEFINED"
    confidentialityRequirement: Optional[CiaRequirementType] = "NOT_DEFINED"
    integrityRequirement: Optional[CiaRequirementType] = "NOT_DEFINED"
    availabilityRequirement: Optional[CiaRequirementType] = "NOT_DEFINED"
    modifiedAttackVector: Optional[ModifiedAttackVectorType] = "NOT_DEFINED"
    modifiedAttackComplexity: Optional[ModifiedAttackComplexityType] = "NOT_DEFINED"
    modifiedAttackRequirements: Optional[ModifiedAttackRequirementsType] = "NOT_DEFINED"
    modifiedPrivilegesRequired: Optional[ModifiedPrivilegesRequiredType] = "NOT_DEFINED"
    modifiedUserInteraction: Optional[ModifiedUserInteractionType] = "NOT_DEFINED"
    modifiedVulnConfidentialityImpact: Optional[ModifiedVulnCiaType] = "NOT_DEFINED"
    modifiedVulnIntegrityImpact: Optional[ModifiedVulnCiaType] = "NOT_DEFINED"
    modifiedVulnAvailabilityImpact: Optional[ModifiedVulnCiaType] = "NOT_DEFINED"
    modifiedSubConfidentialityImpact: Optional[ModifiedSubCType] = "NOT_DEFINED"
    modifiedSubIntegrityImpact: Optional[ModifiedSubIaType] = "NOT_DEFINED"
    modifiedSubAvailabilityImpact: Optional[ModifiedSubIaType] = "NOT_DEFINED"
    Safety: Optional[SafetyType] = "NOT_DEFINED"
    Automatable: Optional[AutomatableType] = "NOT_DEFINED"
    Recovery: Optional[RecoveryType] = "NOT_DEFINED"
    valueDensity: Optional[ValueDensityType] = "NOT_DEFINED"
    vulnerabilityResponseEffort: Optional[VulnerabilityResponseEffortType] = (
        "NOT_DEFINED"
    )
    providerUrgency: Optional[ProviderUrgencyType] = "NOT_DEFINED"



import os
import tarfile

import oras.client
from oras.logger import setup_logger

setup_logger(quiet=True, debug=False)


def download_image(target, outdir):
    """
    Method to download vdb files from a oci registry
    """
    oras_client = oras.client.OrasClient()
    paths_list = oras_client.pull(
        target=target,
        outdir=outdir,
        allowed_media_type=[],
        overwrite=True,
    )
    for apath in paths_list:
        if apath.endswith(".tar.gz") or apath.endswith(".tar.xz"):
            with tarfile.open(apath, "r") as tarf:
                tarf.extractall(path=outdir)
            try:
                os.remove(apath)
            except OSError:
                pass
    return paths_list



# pylint: disable=C0115, C0103, C0301

from enum import Enum
from typing import Annotated, Literal

from pydantic import Field, RootModel


class AttackVectorType(Enum):
    NETWORK = "NETWORK"
    ADJACENT = "ADJACENT"
    LOCAL = "LOCAL"
    PHYSICAL = "PHYSICAL"


class ModifiedAttackVectorType(Enum):
    NETWORK = "NETWORK"
    ADJACENT = "ADJACENT"
    LOCAL = "LOCAL"
    PHYSICAL = "PHYSICAL"
    NOT_DEFINED = "NOT_DEFINED"


class AttackComplexityType(Enum):
    HIGH = "HIGH"
    LOW = "LOW"


class ModifiedAttackComplexityType(Enum):
    HIGH = "HIGH"
    LOW = "LOW"
    NOT_DEFINED = "NOT_DEFINED"


class AttackRequirementsType(Enum):
    NONE = "NONE"
    PRESENT = "PRESENT"


class ModifiedAttackRequirementsType(Enum):
    NONE = "NONE"
    PRESENT = "PRESENT"
    NOT_DEFINED = "NOT_DEFINED"


class PrivilegesRequiredType(Enum):
    HIGH = "HIGH"
    LOW = "LOW"
    NONE = "NONE"


class ModifiedPrivilegesRequiredType(Enum):
    HIGH = "HIGH"
    LOW = "LOW"
    NONE = "NONE"
    NOT_DEFINED = "NOT_DEFINED"


class UserInteractionType(Enum):
    NONE = "NONE"
    PASSIVE = "PASSIVE"
    ACTIVE = "ACTIVE"


class ModifiedUserInteractionType(Enum):
    NONE = "NONE"
    PASSIVE = "PASSIVE"
    ACTIVE = "ACTIVE"
    NOT_DEFINED = "NOT_DEFINED"


class VulnCiaType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    HIGH = "HIGH"


class ModifiedVulnCiaType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class SubCiaType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    HIGH = "HIGH"


class ModifiedSubCType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class ModifiedSubIaType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    HIGH = "HIGH"
    SAFETY = "SAFETY"
    NOT_DEFINED = "NOT_DEFINED"


class ExploitMaturityType(Enum):
    UNREPORTED = "UNREPORTED"
    PROOF_OF_CONCEPT = "PROOF_OF_CONCEPT"
    ATTACKED = "ATTACKED"
    NOT_DEFINED = "NOT_DEFINED"


class CiaRequirementType(Enum):
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class SafetyType(Enum):
    NEGLIGIBLE = "NEGLIGIBLE"
    PRESENT = "PRESENT"
    NOT_DEFINED = "NOT_DEFINED"


class AutomatableType(Enum):
    NO = "NO"
    YES = "YES"
    NOT_DEFINED = "NOT_DEFINED"


class RecoveryType(Enum):
    AUTOMATIC = "AUTOMATIC"
    USER = "USER"
    IRRECOVERABLE = "IRRECOVERABLE"
    NOT_DEFINED = "NOT_DEFINED"


class ValueDensityType(Enum):
    DIFFUSE = "DIFFUSE"
    CONCENTRATED = "CONCENTRATED"
    NOT_DEFINED = "NOT_DEFINED"


class VulnerabilityResponseEffortType(Enum):
    LOW = "LOW"
    MODERATE = "MODERATE"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class ProviderUrgencyType(Enum):
    CLEAR = "CLEAR"
    GREEN = "GREEN"
    AMBER = "AMBER"
    RED = "RED"
    NOT_DEFINED = "NOT_DEFINED"


class ScoreType(Enum):
    number_0_0 = 0.0
    number_0_1 = 0.1
    number_0_2 = 0.2
    number_0_3 = 0.3
    number_0_4 = 0.4
    number_0_5 = 0.5
    number_0_6 = 0.6
    number_0_7 = 0.7
    number_0_8 = 0.8
    number_0_9 = 0.9
    number_1_0 = 1.0
    number_1_1 = 1.1
    number_1_2 = 1.2
    number_1_3 = 1.3
    number_1_4 = 1.4
    number_1_5 = 1.5
    number_1_6 = 1.6
    number_1_7 = 1.7
    number_1_8 = 1.8
    number_1_9 = 1.9
    number_2_0 = 2.0
    number_2_1 = 2.1
    number_2_2 = 2.2
    number_2_3 = 2.3
    number_2_4 = 2.4
    number_2_5 = 2.5
    number_2_6 = 2.6
    number_2_7 = 2.7
    number_2_8 = 2.8
    number_2_9 = 2.9
    number_3_0 = 3.0
    number_3_1 = 3.1
    number_3_2 = 3.2
    number_3_3 = 3.3
    number_3_4 = 3.4
    number_3_5 = 3.5
    number_3_6 = 3.6
    number_3_7 = 3.7
    number_3_8 = 3.8
    number_3_9 = 3.9
    number_4_0 = 4.0
    number_4_1 = 4.1
    number_4_2 = 4.2
    number_4_3 = 4.3
    number_4_4 = 4.4
    number_4_5 = 4.5
    number_4_6 = 4.6
    number_4_7 = 4.7
    number_4_8 = 4.8
    number_4_9 = 4.9
    number_5_0 = 5.0
    number_5_1 = 5.1
    number_5_2 = 5.2
    number_5_3 = 5.3
    number_5_4 = 5.4
    number_5_5 = 5.5
    number_5_6 = 5.6
    number_5_7 = 5.7
    number_5_8 = 5.8
    number_5_9 = 5.9
    number_6_0 = 6.0
    number_6_1 = 6.1
    number_6_2 = 6.2
    number_6_3 = 6.3
    number_6_4 = 6.4
    number_6_5 = 6.5
    number_6_6 = 6.6
    number_6_7 = 6.7
    number_6_8 = 6.8
    number_6_9 = 6.9
    number_7_0 = 7.0
    number_7_1 = 7.1
    number_7_2 = 7.2
    number_7_3 = 7.3
    number_7_4 = 7.4
    number_7_5 = 7.5
    number_7_6 = 7.6
    number_7_7 = 7.7
    number_7_8 = 7.8
    number_7_9 = 7.9
    number_8_0 = 8.0
    number_8_1 = 8.1
    number_8_2 = 8.2
    number_8_3 = 8.3
    number_8_4 = 8.4
    number_8_5 = 8.5
    number_8_6 = 8.6
    number_8_7 = 8.7
    number_8_8 = 8.8
    number_8_9 = 8.9
    number_9_0 = 9.0
    number_9_1 = 9.1
    number_9_2 = 9.2
    number_9_3 = 9.3
    number_9_4 = 9.4
    number_9_5 = 9.5
    number_9_6 = 9.6
    number_9_7 = 9.7
    number_9_8 = 9.8
    number_9_9 = 9.9
    number_10_0 = 10.0


class NoneScoreType(RootModel[float]):
    root: Annotated[float, Field(ge=0.0, le=0.0)]


class LowScoreType(Enum):
    number_0_1 = 0.1
    number_0_2 = 0.2
    number_0_3 = 0.3
    number_0_4 = 0.4
    number_0_5 = 0.5
    number_0_6 = 0.6
    number_0_7 = 0.7
    number_0_8 = 0.8
    number_0_9 = 0.9
    number_1_0 = 1.0
    number_1_1 = 1.1
    number_1_2 = 1.2
    number_1_3 = 1.3
    number_1_4 = 1.4
    number_1_5 = 1.5
    number_1_6 = 1.6
    number_1_7 = 1.7
    number_1_8 = 1.8
    number_1_9 = 1.9
    number_2_0 = 2.0
    number_2_1 = 2.1
    number_2_2 = 2.2
    number_2_3 = 2.3
    number_2_4 = 2.4
    number_2_5 = 2.5
    number_2_6 = 2.6
    number_2_7 = 2.7
    number_2_8 = 2.8
    number_2_9 = 2.9
    number_3_0 = 3.0
    number_3_1 = 3.1
    number_3_2 = 3.2
    number_3_3 = 3.3
    number_3_4 = 3.4
    number_3_5 = 3.5
    number_3_6 = 3.6
    number_3_7 = 3.7
    number_3_8 = 3.8
    number_3_9 = 3.9


class MediumScoreType(Enum):
    number_4_0 = 4.0
    number_4_1 = 4.1
    number_4_2 = 4.2
    number_4_3 = 4.3
    number_4_4 = 4.4
    number_4_5 = 4.5
    number_4_6 = 4.6
    number_4_7 = 4.7
    number_4_8 = 4.8
    number_4_9 = 4.9
    number_5_0 = 5.0
    number_5_1 = 5.1
    number_5_2 = 5.2
    number_5_3 = 5.3
    number_5_4 = 5.4
    number_5_5 = 5.5
    number_5_6 = 5.6
    number_5_7 = 5.7
    number_5_8 = 5.8
    number_5_9 = 5.9
    number_6_0 = 6.0
    number_6_1 = 6.1
    number_6_2 = 6.2
    number_6_3 = 6.3
    number_6_4 = 6.4
    number_6_5 = 6.5
    number_6_6 = 6.6
    number_6_7 = 6.7
    number_6_8 = 6.8
    number_6_9 = 6.9


class HighScoreType(Enum):
    number_7_0 = 7.0
    number_7_1 = 7.1
    number_7_2 = 7.2
    number_7_3 = 7.3
    number_7_4 = 7.4
    number_7_5 = 7.5
    number_7_6 = 7.6
    number_7_7 = 7.7
    number_7_8 = 7.8
    number_7_9 = 7.9
    number_8_0 = 8.0
    number_8_1 = 8.1
    number_8_2 = 8.2
    number_8_3 = 8.3
    number_8_4 = 8.4
    number_8_5 = 8.5
    number_8_6 = 8.6
    number_8_7 = 8.7
    number_8_8 = 8.8
    number_8_9 = 8.9


class CriticalScoreType(Enum):
    number_9_0 = 9.0
    number_9_1 = 9.1
    number_9_2 = 9.2
    number_9_3 = 9.3
    number_9_4 = 9.4
    number_9_5 = 9.5
    number_9_6 = 9.6
    number_9_7 = 9.7
    number_9_8 = 9.8
    number_9_9 = 9.9
    number_10_0 = 10.0


class SeverityType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    CRITICAL = "CRITICAL"


class NoneSeverityType(RootModel[Literal["NONE"]]):
    root: Literal["NONE"]


class LowSeverityType(RootModel[Literal["LOW"]]):
    root: Literal["LOW"]


class MediumSeverityType(RootModel[Literal["MEDIUM"]]):
    root: Literal["MEDIUM"]


class HighSeverityType(RootModel[Literal["HIGH"]]):
    root: Literal["HIGH"]


class CriticalSeverityType(RootModel[Literal["CRITICAL"]]):
    root: Literal["CRITICAL"]


class AttackVectorTypeModel(Enum):
    NETWORK = "NETWORK"
    ADJACENT_NETWORK = "ADJACENT_NETWORK"
    LOCAL = "LOCAL"
    PHYSICAL = "PHYSICAL"


class ModifiedAttackVectorTypeModel(Enum):
    NETWORK = "NETWORK"
    ADJACENT_NETWORK = "ADJACENT_NETWORK"
    LOCAL = "LOCAL"
    PHYSICAL = "PHYSICAL"
    NOT_DEFINED = "NOT_DEFINED"


class UserInteractionTypeModel(Enum):
    NONE = "NONE"
    REQUIRED = "REQUIRED"


class ModifiedUserInteractionTypeModel(Enum):
    NONE = "NONE"
    REQUIRED = "REQUIRED"
    NOT_DEFINED = "NOT_DEFINED"


class ScopeType(Enum):
    UNCHANGED = "UNCHANGED"
    CHANGED = "CHANGED"


class ModifiedScopeType(Enum):
    UNCHANGED = "UNCHANGED"
    CHANGED = "CHANGED"
    NOT_DEFINED = "NOT_DEFINED"


class CiaType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    HIGH = "HIGH"


class ModifiedCiaType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class ExploitCodeMaturityType(Enum):
    UNPROVEN = "UNPROVEN"
    PROOF_OF_CONCEPT = "PROOF_OF_CONCEPT"
    FUNCTIONAL = "FUNCTIONAL"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class RemediationLevelType(Enum):
    OFFICIAL_FIX = "OFFICIAL_FIX"
    TEMPORARY_FIX = "TEMPORARY_FIX"
    WORKAROUND = "WORKAROUND"
    UNAVAILABLE = "UNAVAILABLE"
    NOT_DEFINED = "NOT_DEFINED"


class ConfidenceType(Enum):
    UNKNOWN = "UNKNOWN"
    REASONABLE = "REASONABLE"
    CONFIRMED = "CONFIRMED"
    NOT_DEFINED = "NOT_DEFINED"


class ScoreTypeModel(RootModel[float]):
    root: Annotated[float, Field(ge=0.0, le=10.0)]


class AccessVectorType(Enum):
    NETWORK = "NETWORK"
    ADJACENT_NETWORK = "ADJACENT_NETWORK"
    LOCAL = "LOCAL"


class AccessComplexityType(Enum):
    HIGH = "HIGH"
    MEDIUM = "MEDIUM"
    LOW = "LOW"


class AuthenticationType(Enum):
    MULTIPLE = "MULTIPLE"
    SINGLE = "SINGLE"
    NONE = "NONE"


class CiaTypeModel(Enum):
    NONE = "NONE"
    PARTIAL = "PARTIAL"
    COMPLETE = "COMPLETE"


class ExploitabilityType(Enum):
    UNPROVEN = "UNPROVEN"
    PROOF_OF_CONCEPT = "PROOF_OF_CONCEPT"
    FUNCTIONAL = "FUNCTIONAL"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class ReportConfidenceType(Enum):
    UNCONFIRMED = "UNCONFIRMED"
    UNCORROBORATED = "UNCORROBORATED"
    CONFIRMED = "CONFIRMED"
    NOT_DEFINED = "NOT_DEFINED"


class CollateralDamagePotentialType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    LOW_MEDIUM = "LOW_MEDIUM"
    MEDIUM_HIGH = "MEDIUM_HIGH"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class TargetDistributionType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"



import base64

from rich.markdown import Markdown
from rich.table import Table

from vdb.lib.cve_model import CVE

NEWLINE = "\n"


def add_table_row(table: Table, res: dict, added_row_keys: dict):
    # matched_by is the purl or cpe string
    row_key = f"""{res["matched_by"]}|{res.get("source_data_hash")}"""
    # Filter duplicate rows from getting printed
    if added_row_keys.get(row_key):
        return
    source_data: CVE = res.get("source_data")
    fix_version = res.get("fix_version", "")
    descriptions = []
    cna_container = source_data.root.containers.cna
    affected_functions = set()
    affected_modules = set()
    if cna_container and cna_container.descriptions and cna_container.descriptions.root:
        for adesc in cna_container.descriptions.root:
            description = (
                "\n".join(
                    [
                        base64.b64decode(sm.value).decode("utf-8")
                        for sm in adesc.supportingMedia
                    ]
                )
                if adesc.supportingMedia
                else adesc.value
            )
            description = description.replace("\\n", "\n").replace("\\t", "  ")
            descriptions.append(description)
    if cna_container.affected and cna_container.affected.root:
        for each_affected in cna_container.affected.root:
            if each_affected.programRoutines:
                affected_functions |= {r.name for r in each_affected.programRoutines}
            if each_affected.modules:
                affected_modules |= {m.root for m in each_affected.modules}
    affected_functions = list(affected_functions)
    affected_modules = list(affected_modules)
    affects = ""
    if affected_functions:
        affects = f"## Functions\n- {(NEWLINE + '- ').join(affected_functions)}"
    if affected_modules:
        affects = f"{affects}\n## Modules\n- {(NEWLINE + '- ').join(affected_modules)}"
    matched_by_str = f'{res["matched_by"]}\n{res.get("purl_prefix")}' if not res["matched_by"].startswith("pkg:") else res["matched_by"] 
    table.add_row(
        res.get("cve_id"),
        matched_by_str,
        fix_version,
        Markdown("\n".join(descriptions), justify="left", hyperlinks=True),
        Markdown(affects, justify="left"),
    )
    added_row_keys[row_key] = True
